package org.blueoxygen.cimande.persistence.hibernate;

import java.io.Serializable;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.blueoxygen.cimande.persistence.GenericPersistenceManager;
import org.blueoxygen.cimande.persistence.PersistenceException;
import org.hibernate.Criteria;
import org.hibernate.Hibernate;
import org.hibernate.HibernateException;
import org.hibernate.ObjectDeletedException;
import org.hibernate.ObjectNotFoundException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.criterion.Projections;
import org.hibernate.type.Type;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

@Repository
@Transactional(readOnly = true)
public class GenericHibernatePersistenceManager implements
		GenericPersistenceManager {

	@Autowired
	private SessionFactory sessionFactory;

	@Transactional(readOnly = true)
	public long getRowCount(Class clazz) {
		Criteria criteria = sessionFactory.getCurrentSession().createCriteria(
				clazz);
		criteria.setProjection(Projections.rowCount());

		return new Long(criteria.uniqueResult() + "");
	}

	@Transactional
	public void save(Object objectToSave) {
		try {
			sessionFactory.getCurrentSession().saveOrUpdate(objectToSave);
		} catch (Exception e) {
			throw new PersistenceException(e);
		}
	}

	@Transactional
	public void persist(Serializable objectToSave) {
		sessionFactory.getCurrentSession().persist(objectToSave);
	}

	@Transactional
	public void merge(Serializable objectToSave) {
		sessionFactory.getCurrentSession().merge(objectToSave);
	}

	@Transactional
	public void remove(Object objectToRemove) {
		try {
			sessionFactory.getCurrentSession().delete(objectToRemove);
		} catch (Exception e) {
			throw new PersistenceException(e);
		}
	}

	@Transactional(readOnly = true)
	public <T> T getById(Class<T> type, long id) {
		try {
			// return session.get(, new Long(id));
			return (T) sessionFactory.getCurrentSession().get(type,
					new Long(id));
		} catch (ObjectDeletedException e) {
			return null;
		} catch (ObjectNotFoundException e) {
			return null;
		} catch (Exception e) {
			throw new PersistenceException(e);
		}
	}

	@Transactional(readOnly = true)
	public <T> T getByUniqueField(Class<T> type, Object pk, String fieldName) {
		Query query = sessionFactory.getCurrentSession().createQuery(
				"SELECT a FROM " + type.getName() + " a WHERE a." + fieldName
						+ "=:pk");
		query.setParameter("pk", pk);
		try {
			return (T) query.uniqueResult();
		} catch (ObjectNotFoundException e) {
			return null;
		} catch (HibernateException e) {
			throw new PersistenceException(e);
		}
	}

	public <T> List<T> findAll(Class<T> type) {
		return findAllSorted(type, null);
	}

	@Transactional(readOnly = true)
	public <T> List<T> findAllSorted(Class<T> type, String sortField) {
		try {
			String query = "FROM " + type.getSimpleName() + " as result";
			if (sortField != null)
				query += " ORDER BY LOWER(result." + sortField + ")";
			return sessionFactory.getCurrentSession().createQuery(query).list();
		} catch (Exception e) {
			throw new PersistenceException(e);
		}
	}

	@Transactional(readOnly = true)
	public <T> List<T> findAllSortedDirected(Class<T> type, String sortField,
			String direction) {
		try {
			String query = "FROM " + type.getSimpleName() + " as result";
			if (sortField != null)
				query += " ORDER BY LOWER(result." + sortField + ")";
			if (direction != null) {
				if (direction.equalsIgnoreCase("asc")) {
					query += " asc";
				}
				if (direction.equalsIgnoreCase("desc")) {
					query += " desc";
				}
			}
			return sessionFactory.getCurrentSession().createQuery(query).list();
		} catch (Exception e) {
			throw new PersistenceException(e);
		}
	}

	@Transactional(readOnly = true)
	public List find(String query, Object[] parameters, Class[] parameterTypes) {
		Session session = null;
		List results = null;
		try {
			session = sessionFactory.getCurrentSession();
			if ((parameterTypes != null) || (parameters != null)) {
				Type[] hibernate_parameter_types = getHibernatedParameterTypes(parameterTypes);
				results = session
						.createQuery(query)
						.setParameters(parameters,
								getHibernatedParameterTypes(parameterTypes))
						.list();
			} else {
				results = session.createQuery(query).list();
			}
			return results;
		} catch (Exception e) {
			throw new PersistenceException(e);
		}
	}

	@Transactional(readOnly = true)
	public List getList(String query, Object[] parameters,
			Class[] parameterTypes) {

		// Session session = null;
		List results = new ArrayList();

		try {
			Session session = sessionFactory.getCurrentSession();

			if ((parameterTypes != null) || (parameters != null)) {
				Type[] hibernate_parameter_types = getHibernatedParameterTypes(parameterTypes);
				Query q = session.createQuery(query).setParameters(parameters,
						hibernate_parameter_types);
				results = q.list();
			} else {
				results = session.createQuery(query).list();
			}

			return results;
		} catch (Exception e) {
			throw new PersistenceException(e);
		}
	}

	@Transactional(readOnly = true)
	public <T> T getById(Class<T> aClass, String id) {
		try {
			return (T) sessionFactory.getCurrentSession().get(aClass,
					new String(id));
		} catch (ObjectDeletedException e) {
			return null;
		} catch (ObjectNotFoundException e) {
			return null;
		} catch (Exception e) {
			throw new PersistenceException(e);
		}
	}

	private static Type[] getHibernatedParameterTypes(Class[] types) {
		if (types == null) {
			return null;
		}

		Type[] hib_types = new Type[types.length];

		for (int i = 0; i < types.length; i++) {
			Class type = types[i];

			hib_types[i] = (Type) classToHibernateTypeMap.get(type);
		}

		return hib_types;
	}

	private static Map classToHibernateTypeMap = new HashMap();

	static {
		classToHibernateTypeMap.put(Boolean.class, Hibernate.BIG_DECIMAL);
		classToHibernateTypeMap.put(Boolean.class, Hibernate.BOOLEAN);
		classToHibernateTypeMap.put(Byte.class, Hibernate.BYTE);
		classToHibernateTypeMap.put(Character.class, Hibernate.CHARACTER);
		classToHibernateTypeMap.put(Date.class, Hibernate.DATE);
		classToHibernateTypeMap.put(Double.class, Hibernate.DOUBLE);
		classToHibernateTypeMap.put(Float.class, Hibernate.FLOAT);
		classToHibernateTypeMap.put(Integer.class, Hibernate.INTEGER);
		classToHibernateTypeMap.put(Long.class, Hibernate.LONG);
		classToHibernateTypeMap.put(Short.class, Hibernate.SHORT);
		classToHibernateTypeMap.put(String.class, Hibernate.STRING);
		classToHibernateTypeMap.put(Timestamp.class, Hibernate.TIMESTAMP);
	}

}
